__global__ static void hst_cuda_kernel1(float *cache, int num_proj, int num_bins, float *d_SLICE, float apos_off_x, float apos_off_y, int batch) {
    const int tidx = threadIdx.x;
    const int tidy = threadIdx.y;

    const int bidx = blockIdx.x * BLOCK_SIZE_X;
    const int bidy = batch + blockIdx.y * BLOCK_SIZE_Y;

    const int bid = blockIdx.y * gridDim.x + blockIdx.x;
    const int width = 2 * BLOCK_SIZE_Y * gridDim.x * gridDim.y;

    float4 all1, all2;

    const float bx = bidx + apos_off_x;
    const float by = bidy + apos_off_y;

    int minh1, minh2;

    const int prj = tidx >> 3;
    const int pos = BLOCK_SIZE_X + (tidx & 7);

//    const int prj = tidx&1;
//    const int pos = BLOCK_SIZE_X + (tidx >> 1);

    const int tidy2 = tidy << 1;
    

/*
//#pragma unroll 1
    for (int proje=0; proje<num_proj; proje += 2 * BLOCK_SIZE_Y) {
	const int proj = proje + tidy2;

        all1 = c_all[proj];
        all2 = c_all[proj + 1];
	minh1 = (int)(all1.z + bx * all1.x - by * all1.y + all1.w);
	minh2 = (int)(all2.z + bx * all2.x - by * all2.y + all2.w);

	cache[(proj)*width + bid * 32 + tidx] = tex2D(tex_projes, minh1 + tidx, proj);
	cache[(proj+1)*width + bid * 32 + tidx] = tex2D(tex_projes, minh2 + tidx, proj + 1);
	cache[(proj+prj)*width + bid * 32 + pos] = tex2D(tex_projes, (prj?minh2:minh1) + pos, proj + prj);
    }
*/

//#pragma unroll 16
    for (int proje=0; proje<num_proj; proje += BLOCK_SIZE_Y) {
	const int proj = proje + tidy;

        all1 = c_all[proj];
        all2 = c_all[proj + 1];
	minh1 = (int)(all1.z + bx * all1.x - by * all1.y + all1.w);
	minh2 = (int)(all2.z + bx * all2.x - by * all2.y + all2.w);

	if (tidx < 12) {
	    cache[proj*width + bid * 32 + tidx] = tex2D(tex_projes, minh1 + tidx, proj);
	    cache[proj*width + bid * 32 + 12 + tidx] = tex2D(tex_projes, minh1 + 12 + tidx, proj);
	}
    }

}



#define OVERSAMPLING 4
__global__ static void hst_cuda_kernel2(int num_proj, int num_bins, float *d_SLICE, float apos_off_x, float apos_off_y, int batch) {
    int i;
    int proj;
    
    const int tidx = threadIdx.x;
    const int tidy = threadIdx.y;

    const int bidx = blockIdx.x * BLOCK_SIZE_X;
    const int bidy = batch + blockIdx.y * BLOCK_SIZE_Y;

    const int idx = bidx + threadIdx.x;
    const int idy = bidy + threadIdx.y;

    const int bid = blockIdx.y * gridDim.x + blockIdx.x;
    const int width = 2 * BLOCK_SIZE_Y * gridDim.x * gridDim.y;

    float4 all;

    float res = 0;

    const float bx = bidx + apos_off_x;
    const float by = bidy + apos_off_y;

    const float x = bx + tidx;
    const float y = by + tidy;

    float h;
    int minh;

    __shared__  float   cache_subh[BLOCK_SIZE_X][BLOCK_SIZE_Y + 1];
    __shared__  float   cache_sin[BLOCK_SIZE_Y];
    __shared__  float   cache[BLOCK_SIZE_Y][2 * OVERSAMPLING * BLOCK_SIZE_X];
    
//#pragma unroll 8
    for (int proje=0; proje<num_proj; proje += BLOCK_SIZE_Y) {
	int proj = proje + tidy;

        all = c_all[proj];
	minh = (int)(all.z + bx * all.x - by * all.y + all.w);
	cache_subh[tidx][tidy] = OVERSAMPLING * (all.z + x * all.x - minh);
	if (tidx == 0) cache_sin[tidy] = OVERSAMPLING * all.y;

#pragma unroll 6
	for (i = 0; i < 6/*2 * OVERSAMPLING*/; i++) {
	    int pos = i * BLOCK_SIZE_X + tidx;
	    cache[tidy][pos] = tex2D(tex_linear, bid * 32 + ((float)pos)/OVERSAMPLING , proje + tidy);
	}


	__syncthreads();

	
#pragma unroll 16
	for (i = 0; i < BLOCK_SIZE_Y; i++) {
	    minh = cache_subh[tidx][i] - y * cache_sin[i];
	    res += cache[i][minh];
	}


	__syncthreads();

    }

    d_SLICE[ BLOCK_SIZE_X*gridDim.x*idy + idx ] = res;
}



/*
__global__ static void hst_cuda_kernel2(int num_proj, int num_bins, float *d_SLICE, float apos_off_x, float apos_off_y, int batch) {
    const int tidx = threadIdx.x;
    const int tidy = threadIdx.y;

    const int bid = blockIdx.y * gridDim.x + blockIdx.x;
    const int width = 2 * BLOCK_SIZE_Y * gridDim.x * gridDim.y;

    const int bidx = blockIdx.x * BLOCK_SIZE_X;
    const int bidy = batch + blockIdx.y * BLOCK_SIZE_Y;

    const int idx = bidx + threadIdx.x;
    const int idy = bidy + threadIdx.y;

    float4 all1, all2;

    float res = 0;

    const float bx = bidx + apos_off_x;
    const float by = bidy + apos_off_y;

    const float x = bx + tidx;
    const float y = by + tidy;

    float minh;
    int minh1, minh2;

    __shared__  float   cache_subh[BLOCK_SIZE_X][2 * BLOCK_SIZE_Y + 1];
    __shared__  float   cache_sin[2 * BLOCK_SIZE_Y];

    const int prj = tidx >> 3;
    const int pos = BLOCK_SIZE_X + (tidx & 7);
//	int prj = tidx&1;
//	int pos = BLOCK_SIZE_X + (tidx >> 1);

    const int tidy2 = tidy << 1;
    const int tidy21 = tidy2 + 1;
    const int tidy2p = tidy2 + prj;

//#pragma unroll 1
    for (int proje=0; proje<num_proj; proje += 2 * BLOCK_SIZE_Y) {
	int proj = proje + tidy2;
	
        all1 = c_all[proj];
        all2 = c_all[proj + 1];
	minh1 = (int)(all1.z + bx * all1.x - by * all1.y + all1.w);
	minh2 = (int)(all2.z + bx * all2.x - by * all2.y + all2.w);
		
	cache_subh[tidx][tidy2] = all1.z + x * all1.x - minh1;
	cache_subh[tidx][tidy21] = all2.z + x * all2.x - minh2;
	
	if (tidx == 0) {
	    cache_sin[tidy2] = all1.y;
	    cache_sin[tidy21] = all2.y;
	}

	__syncthreads();


#pragma unroll 16
	for (int i = 0; i < 2*BLOCK_SIZE_Y; i++) {
	    int j = 2 * i;
//    	    const float4 all = c_all[proje];
//	    minh = (int)(all.z + bx * all.x - by * all.y + all.w);
//    	    h = all.z + x * all.x - y * all.y - minh;

	    minh = cache_subh[tidx][i] - y * cache_sin[i];
	    res += tex2D(tex_linear, bid * 32 + minh, proje + i);
//    	    res += tex1D(tex_linear, (proje + i) * width + bid * 32 + minh);
//          res += cache[(proje + i) * width + bid * 32 + (int)minh];

	}

	__syncthreads();

    }

    d_SLICE[ BLOCK_SIZE_X*gridDim.x*idy + idx ] = res;
}
*/
