#undef subh_cache
#define subh_cache(subh) cache[(int)(subh)]

#define PPT 6

__kernel void hst_backproject_test_oversample4_1(const int num_proj, 
    const int num_bins, 
    const float off_x, 
    const float off_y, 
    __read_only image2d_t sinogram,
    __global float *slice,
    __constant float4 *c_all, 
    __local float *cache
    ) {

    const int tidx = get_local_id(0);
    const int tidy = get_local_id(1);

    const int bidx = PPT * get_group_id(0) * BLOCK_SIZE;
    const int bidy = PPT * (get_global_id(1) - get_local_id(1) - get_group_id(1) * BLOCK_SIZE);

    const float bx = bidx + off_x;
    const float by = bidy + off_y;

/*
    const int sbidx = tidy % 4;
    const int sbidy = tidy / 4;
    
    const int stidx = tidx % 4;
    const int stidy = tidx / 4;

    const int sidx = (sbidx * 4 + stidx);
    const int sidy = (sbidy * 4 + stidy);
*/

    const int sidx = tidx;
    const int sidy = tidy - BLOCK_SIZE;


    const float sx = sidx;
    const float sy = sidy;

    const int idx = bidx + sidx;
    const int idy = bidy + sidy;

    const int slice_width = PPT * get_global_size(0);

    int buf = 0;
    int offset = 5 * BLOCK_SIZE * BLOCK_SIZE;
    float4 all = c_all[0];
    
    if (tidy < BLOCK_SIZE) {
        float minh = floor(all.z + bx * all.x - by * all.y + all.w);

#pragma unroll 5
        for (int i = 0; i < 5; i++) {
	    int pos = i * BLOCK_SIZE * BLOCK_SIZE + tidy * BLOCK_SIZE + tidx;
	    cache[pos] = read_imagef(sinogram, volumeSampler, (float2)(minh + 0.25f*pos, .5f)).x;
	}
    }
    
    float res[PPT][PPT] = {0};
    for (int proj = 0; proj < num_proj; proj++) {
        if (get_local_id(1) >= BLOCK_SIZE) {
            all = -4 * all; 
            all.x = -all.x;

            float subh = mad(sx, all.x, mad(sy, all.y, all.w));


#pragma unroll 6
            for (int i = 0; i < PPT; i++) {
#pragma unroll 6
                for (int j = 0; j < PPT; j++) {
                    res[i][j] += subh_cache(buf * offset + subh + (i * BLOCK_SIZE) * all.y + (j * BLOCK_SIZE) * all.x);
                }
            }
        }
        
        buf = !buf;
        if (proj == num_proj) break;

        all = c_all[proj + 1];
        if (tidy < BLOCK_SIZE) {
            float minh = floor(all.z + bx * all.x - by * all.y + all.w);

#pragma unroll 5
            for (int i = 0; i < 5; i++) {
	        int pos = i * BLOCK_SIZE * BLOCK_SIZE + tidy * BLOCK_SIZE + tidx;
	        cache[buf * offset + pos] = read_imagef(sinogram, volumeSampler, (float2)(minh + 0.25f*pos, proj + 1.5f)).x;
	    }
        }

        barrier(CLK_LOCAL_MEM_FENCE);
    }

    if (get_local_id(1) >= BLOCK_SIZE) {
#pragma unroll 6
    for (int i = 0; i < PPT; i++) {
#pragma unroll 6
        for (int j = 0; j < PPT; j++) {
            slice[(idy + BLOCK_SIZE * i) * slice_width +  idx + BLOCK_SIZE * j] = res[i][j];
        }
    }
    }
}


__kernel void hst_backproject_test_oversample4_2(const int num_proj, 
    const int num_bins, 
    const float off_x, 
    const float off_y, 
    __read_only image2d_t sinogram,
    __global float *slice,
    __constant float4 *c_all, 
    __local float *shared
    ) {

    const int tidx = get_local_id(0);
    const int tidy = get_local_id(1);

    const int sidx = tidx;
    const int sidy = tidy;

    const int idx = PPT * get_group_id(0) * BLOCK_SIZE + sidx;
    const int idy = PPT * (get_global_id(1) - get_local_id(1) - get_group_id(1) * BLOCK_SIZE) + sidy;

    const float x = idx + off_x;
    const float y = idy + off_y;

    const int slice_width = PPT * get_global_size(0);

    float res[PPT][PPT] = {0};
    for (int proj = 0; proj < num_proj; proj++) {
        float4 all = c_all[proj];
        all.y = - all.y;
        float subh = mad(x, all.x, mad(y, all.y, all.z));

#pragma unroll 6
        for (int i = 0; i < PPT; i++) {
#pragma unroll 6
            for (int j = 0; j < PPT; j++) {
                res[i][j] += read_imagef(sinogram, volumeSampler, (float2)(subh + i * BLOCK_SIZE * all.y + j * BLOCK_SIZE * all.x, proj + .5f)).x;
            }
        }
    }

#pragma unroll 6
    for (int i = 0; i < PPT; i++) {
#pragma unroll 6
        for (int j = 0; j < PPT; j++) {
            slice[(idy + BLOCK_SIZE * i) * slice_width +  idx + BLOCK_SIZE * j] = res[i][j];
        }
    }
}



__kernel void hst_backproject_test_oversample4(const int num_proj, 
    const int num_bins, 
    const float off_x, 
    const float off_y, 
    __read_only image2d_t sinogram,
    __global float *slice,
    __constant float4 *c_all, 
    __local float *cache) {

    const int mode = 4;//(get_group_id(1) * get_num_groups(0) + get_group_id(0)) % 8;
    if (mode < 3) {
    if (get_local_id(1) < BLOCK_SIZE) {
        hst_backproject_test_oversample4_2(num_proj, num_bins, off_x, off_y, sinogram, slice, c_all, cache);
    }
    } else {
        hst_backproject_test_oversample4_1(num_proj, num_bins, off_x, off_y, sinogram, slice, c_all, cache);
    }
}

