/* 
 Semi-direct linear interpolation (without use of local arrays) with approximation 
 expanded to use float4 types. No speed gain
*/

#define subh_cache(subh) cache[(int)(subh)]
__kernel void hst_backproject_direct_oversample4(const int num_proj, 
    const int num_bins, 
    const float off_x, 
    const float off_y, 
    __read_only image2d_t sinogram,
    __global float *slice,
    __constant float4 *c_all, 
    __local float *shared
    ) {

    __local float2 *cache = (__local float2 *)shared;						// cache[BLOCK_SIZE_Y][12 * BLOCK_SIZE_X + 1];

    const int tidx = get_local_id(0);
    const int tidy = get_local_id(1);

    const int bidx = 2 * get_group_id(0) * BLOCK_SIZE;
    const int bidy = 2 * get_group_id(1) * BLOCK_SIZE;

    const float bx = bidx + off_x;
    const float by = bidy + off_y;

        // This is subblocks within block (big block is 16x16, subblock is 4x4 completely contained in a single wrap)
    const int sbidx = tidy % 4;
    const int sbidy = tidy / 4;
    
    const int stidx = tidx % 4;
    const int stidy = tidx / 4;
    
    const int sidx = - (sbidx * 8 + 2 * stidx);
    const int sidy = sbidy * 8 + 2 * stidy;

    const int idx = bidx - sidx;
    const int idy = bidy + sidy;


    
    const int slice_width = 2 * get_global_size(0);

    float4 res = {0.f, 0.f, 0.f, 0.f};
    
    for (int proje=0; proje<num_proj; proje += BLOCK_SIZE) {
	int proj = proje + tidy;
        float4 all = c_all[proj];
	float minh = floor(all.z + bx * all.x - by * all.y + all.w);

#pragma unroll 12
	for (int i = 0; i < 3; i++) {
	    int pos = i * BLOCK_SIZE + tidx;
	    cache[tidy * (3 * BLOCK_SIZE + 1) + pos].x = read_imagef(sinogram, volumeSampler, (float2)(minh + pos, proj + 0.5f)).x;
	}

        barrier(CLK_LOCAL_MEM_FENCE);

#pragma unroll 3
	for (int i = 0; i < 3; i++) {
	    int pos = i * BLOCK_SIZE + tidx;
	    cache[tidy * (3 * BLOCK_SIZE + 1) + pos].y = cache[tidy * (3 * BLOCK_SIZE + 1) + pos + 1].x - cache[tidy * (3 * BLOCK_SIZE + 1) + pos].x;
	}

        barrier(CLK_LOCAL_MEM_FENCE);

#pragma unroll 4
	for (int i = 0; i < 4; i++) {
	    int prj = mul24(4, i) + proje;   //mad24(4, i, proje); // insignificantly faster on 59xx, but signifcantly slower on 79xx
            float2 v;
            float4 v1, v2;
            float4 h;
            float4 ih;
            float4 diff;

            float4 i0123 = {0, 1, 2, 3};
	    float4 i4 = 49.f * (4 * i + i0123);
            
//            {pos, pos + 1, pos + 2, pos + 3};

            float4 all1 = c_all[prj];
            float4 all2 = c_all[prj + 1];
            float4 all3 = c_all[prj + 2];
            float4 all4 = c_all[prj + 3];


            float4 allx = {-all1.x, -all2.x, -all3.x, -all4.x};
            float4 ally = {-all1.y, -all2.y, -all3.y, -all4.y};
            float4 allw = {-all1.w, -all2.w, -all3.w, -all4.w};

            float4 hx = i4 + mad(sidx, allx, mad(sidy, ally, allw));
            float4 hy = mad(-1.f, allx, hx);
            float4 hz = hx + ally;
            float4 hw = hy + ally;
{            
            float4 h = {hx.x, hy.x, hz.x, hw.x};
            ih = floor(h);
            diff = h - ih;
            
            v = cache[(int)ih.x]; v1.x = v.x; v2.x = v.y;
            v = cache[(int)ih.y]; v1.y = v.x; v2.y = v.y;
            v = cache[(int)ih.z]; v1.z = v.x; v2.z = v.y;
            v = cache[(int)ih.w]; v1.w = v.x; v2.w = v.y;
            res += v1  + diff * v2;
}
{
            float4 h = {hx.y, hy.y, hz.y, hw.y};
            ih = floor(h);
            diff = h - ih;
            
            v = cache[(int)ih.x]; v1.x = v.x; v2.x = v.y;
            v = cache[(int)ih.y]; v1.y = v.x; v2.y = v.y;
            v = cache[(int)ih.z]; v1.z = v.x; v2.z = v.y;
            v = cache[(int)ih.w]; v1.w = v.x; v2.w = v.y;
            res += v1  + diff * v2;
}
{
            float4 h = {hx.z, hy.z, hz.z, hw.z};
            ih = floor(h);
            diff = h - ih;
            
            v = cache[(int)ih.x]; v1.x = v.x; v2.x = v.y;
            v = cache[(int)ih.y]; v1.y = v.x; v2.y = v.y;
            v = cache[(int)ih.z]; v1.z = v.x; v2.z = v.y;
            v = cache[(int)ih.w]; v1.w = v.x; v2.w = v.y;
            res += v1  + diff * v2;
}
{
            float4 h = {hx.w, hy.w, hz.w, hw.w};
            ih = floor(h);
            diff = h - ih;
            
            v = cache[(int)ih.x]; v1.x = v.x; v2.x = v.y;
            v = cache[(int)ih.y]; v1.y = v.x; v2.y = v.y;
            v = cache[(int)ih.z]; v1.z = v.x; v2.z = v.y;
            v = cache[(int)ih.w]; v1.w = v.x; v2.w = v.y;
            res += v1  + diff * v2;
}
	}

        barrier(CLK_LOCAL_MEM_FENCE);
    }

    slice[slice_width * (idy    ) + idx] = res.x;
    slice[slice_width * (idy    ) + idx + 1] = res.y;
    slice[slice_width * (idy + 1) + idx    ] = res.z;
    slice[slice_width * (idy + 1) + idx + 1] = res.w;
}
