/*
 Iterating only a single projection (requires +4 Y blocks in the management code). It is quite
 slow on Fermi 63 vs. 45. No enlargement of the block tried because in the best case it would
 just compensate the slow down
*/


#undef subh_cache
#define subh_cache(subh) cache[(int)(subh)]
__kernel void hst_backproject_test_oversample4(const int num_proj, 
    const int num_bins, 
    const float off_x, 
    const float off_y, 
    __read_only image2d_t sinogram,
    __global float *slice,
    __constant float4 *c_all, 
    __local float *shared
    ) {

    __local float *cache = shared;						// cache[BLOCK_SIZE_Y][12 * BLOCK_SIZE_X + 1];

    const int tidx = get_local_id(0);
    const int tidy = get_local_id(1);

    const int bidx = 2 * get_group_id(0) * BLOCK_SIZE;
    const int bidy = 2 * get_group_id(1) * BLOCK_SIZE;

    const float bx = bidx + off_x;
    const float by = bidy + off_y;


    const int rtidy = (tidy - 4);// * 2 + tidx / 8;
    
        // This is subblocks within block (big block is 16x16, subblock is 4x4 completely contained in a single wrap)
    const int sbidx = rtidy % 4;
    const int sbidy = rtidy / 4;
    
    const int stidx = tidx % 4;
    const int stidy = tidx / 4;
    
    const int sidx = - (sbidx * 8 + 2 * stidx);
    const int sidy = sbidy * 8 + 2 * stidy;

    const int idx = bidx - sidx;
    const int idy = bidy + sidy;
    
    const int slice_width = 2 * get_global_size(0);

    float4 res = {0.f, 0.f, 0.f, 0.f};

    int buf = 0;
    float4 all;
    float minh;
    

        if (tidy < 4) {
            all = c_all[tidy];
	    minh = all.z + bx * all.x - by * all.y + all.w;

#pragma unroll 3
	    for (int i = 0; i < 3; i++) {
	        int pos = (4 * i + tidy) * BLOCK_SIZE + tidx;
	        cache[pos] = read_imagef(sinogram, volumeSampler, (float2)(minh + 0.25f*pos, 0.5f)).x;
	    }
	}

    barrier(CLK_LOCAL_MEM_FENCE);

#pragma unroll 16
    for (int proj = 0; proj < num_proj; proj++) {
        if (tidy > 3) {
            all = -4 * c_all[proj];

                //float subh = (-sidx * all.x - sidy * all.y) - all.w;
            float subh = mad(193.f, buf, mad(sidx, all.x, mad(sidy, all.y, all.w)));
            
	    res.x += subh_cache(subh);
	    res.y += subh_cache(mad(0.f, all.y, subh) - all.x);
	    res.z += subh_cache(mad(1.f, all.y, subh));
	    res.w += subh_cache(mad(1.f, all.y, subh) - all.x);
        }
        
        if (proj >= num_proj) break;
        buf = !buf;

        if (tidy < 4) {
            all = c_all[proj + 1];
	    minh = all.z + bx * all.x - by * all.y + all.w;

#pragma unroll 3
	    for (int i = 0; i < 3; i++) {
	        int pos = (4 * i + tidy) * BLOCK_SIZE + tidx;
	        cache[buf * (12 * BLOCK_SIZE + 1) + pos] = read_imagef(sinogram, volumeSampler, (float2)(minh + 0.25f*pos, proj + 1.5f)).x;
	    }
	}
        
        barrier(CLK_LOCAL_MEM_FENCE);
    }

    if (tidy > 3) {
        slice[slice_width * (idy    ) + idx] = res.x;
        slice[slice_width * (idy + 0) + idx + 1] = res.y;
        slice[slice_width * (idy + 1) + idx] = res.z;
        slice[slice_width * (idy + 1) + idx + 1] = res.w;
    }

}
