
#undef subh_cache
#define subh_cache(subh) cache[(int)(subh)]

#define PPT 6

__kernel void hst_backproject_test_oversample4(const int num_proj, 
    const int num_bins, 
    const float off_x, 
    const float off_y, 
    __read_only image2d_t sinogram,
    __global float *slice,
    __constant float4 *c_all, 
    __local float *shared
    ) {

    __local float *cache = shared;						// cache[BLOCK_SIZE_Y][12 * BLOCK_SIZE_X + 1];

    const int tidx = get_local_id(0);
    const int tidy = get_local_id(1);

    const int bidx = get_group_id(0) * BLOCK_SIZE;
    const int bidy = get_group_id(1) * BLOCK_SIZE;

    const float bx = PPT * bidx + off_x;
    const float by = PPT * bidy + off_y;


    const int rtidy = tidy;// * 2 + tidx / 8;
    
        // This is subblocks within block (big block is 16x16, subblock is 4x4 completely contained in a single wrap)
    const int sbidx = rtidy % 4;
    const int sbidy = rtidy / 4;
    
    const int stidx = tidx % 4;
    const int stidy = tidx / 4;
    
    const int sidx = (sbidx * 4 + stidx);
    const int sidy = (sbidy * 4 +  stidy);

    const int sx = PPT * sidx;
    const int sy = PPT * sidy;

    const int idx = bidx + sidx;
    const int idy = bidy + sidy;
    
    const float x = bx + sx;
    const float y = by + sy;
    
    const int slice_width = get_global_size(0);

    float res[PPT][PPT] = {0};


    for (int proj = 0; proj < num_proj; proj++) {

        float4 all = c_all[proj];
        int minh = floor(all.z + bx * all.x - by * all.y + all.w);
        float subh = 4 * (all.z + x * all.x - y * all.y - minh);

#pragma unroll 2
        for (int i = 0; i < 2; i++) {
	    int pos = i * BLOCK_SIZE * BLOCK_SIZE + tidy * BLOCK_SIZE + tidx;
	    cache[pos] = read_imagef(sinogram, volumeSampler, (float2)(minh + 0.25f*pos, proj + .5f)).x;
	}

        barrier(CLK_LOCAL_MEM_FENCE);

//        all = -4 * all; 
//        float subh = mad(sx, all.x, mad(sy, all.y, all.w));
        
#pragma unroll 6
        for (int i = 0; i < PPT; i++) {
#pragma unroll 6
            for (int j = 0; j < PPT; j++) {
                res[i][j] += subh_cache(subh + 4 * j * all.x - 4 * i * all.y);
            }
        }

        barrier(CLK_LOCAL_MEM_FENCE);
    }


#pragma unroll 6
    for (int i = 0; i < PPT; i++) {
#pragma unroll 6
        for (int j = 0; j < PPT; j++) {
            slice[(idy * PPT + i) * PPT * slice_width +  idx * PPT + j] = res[i][j];
        }
    }

}

