/*
 This is attempt to make float4 arithmetic to simplify task for AMD
 VLIW engine. Actually, it gives very insignificant speed-up
*/

#define to_int4(h) {(int)(h.x), (int)(h.y), (int)(h.z), (int)(h.w)}
#define subh_cache(subh) cache[subh]

__kernel void hst_backproject_direct_oversample4(const int num_proj, 
    const int num_bins, 
    const float off_x, 
    const float off_y, 
    __read_only image2d_t sinogram,
    __global float *slice,
    __constant float4 *c_all, 
    __local float *shared
    ) {

    __local float *cache = shared;						// cache[BLOCK_SIZE_Y][12 * BLOCK_SIZE_X];

    const int tidx = get_local_id(0);
    const int tidy = get_local_id(1);

    const int tidx2 = 2 * get_local_id(0);
    const int tidy2 = 2 * get_local_id(1);

    const int bidx = 2 * get_group_id(0) * BLOCK_SIZE;
    const int bidy = 2 * get_group_id(1) * BLOCK_SIZE;

        // This is subblocks within block (big block is 16x16, subblock is 4x4 completely contained in a single wrap)
    const int sbidx = tidy % 4;
    const int sbidy = tidy / 4;
    
    const int stidx = tidx % 4;
    const int stidy = tidx / 4;
    
    const int sidx = sbidx * 4 + stidx;
    const int sidy = sbidy * 4 + stidy;
    
    const int sidx2 = -2 * sidx;
    const int sidy2 = 2 * sidy;

    const int idx = bidx - sidx2;//2 * get_global_id(0);
    const int idy = bidy + sidy2;//2 * get_global_id(1);

    const float bx = bidx + off_x;
    const float by = bidy + off_y;

//    const float x = idx + off_x;
//    const float y = idy + off_y;
    
    const int slice_width = 2 * get_global_size(0);

    float4 res = {0.f, 0.f, 0.f, 0.f};
    
    for (int proje=0; proje<num_proj; proje += BLOCK_SIZE) {
        {
	    int proj = proje + tidy;
            float4 all = c_all[proj];
	    float minh = all.z + bx * all.x - by * all.y + all.w;

#pragma unroll 12
	    for (int i = 0; i < 12; i++) {
	        int pos = i * BLOCK_SIZE + tidx;
	        cache[tidy * 12 * BLOCK_SIZE + pos] = read_imagef(sinogram, volumeSampler, (float2)(minh + 0.25f*pos, proj + 0.5f)).x;
	    }
	}

        barrier(CLK_LOCAL_MEM_FENCE);

#pragma unroll 4
	for (int i = 0; i < 4; i++) {
	    int prj = mul24(4, i) + proje;   //mad24(4, i, proje); // insignificantly faster on 59xx, but signifcantly slower on 79xx

	    float pos = 768.f * i;
            float4 i4 = {pos, pos + 1*192, pos + 2*192, pos + 3*192};

            float4 all1 = c_all[prj];
            float4 all2 = c_all[prj + 1];
            float4 all3 = c_all[prj + 2];
            float4 all4 = c_all[prj + 3];

            float4 allx = {-4 * all1.x, -4 * all2.x, -4 * all3.x, -4 * all4.x};
            float4 ally = {-4 * all1.y, -4 * all2.y, -4 * all3.y, -4 * all4.y};
            float4 allw = {-4 * all1.w, -4 * all2.w, -4 * all3.w, -4 * all4.w};

//            float subh = 192 * i + (sidx2 * all.x + sidy2 * all.y) + all.w;
//            float subhx = subh - all.x;

            float4 subh = i4 + mad(sidx2, allx, mad(sidy2, ally, allw));
            float4 subhx = mad(-1.f, allx, subh);
            float4 subhy = subh + ally;
            float4 subhxy = subhx + ally;

            int4 h = to_int4(subh);
            int4 hx = to_int4(subhx);
            int4 hy = to_int4(subhy);
            int4 hxy = to_int4(subhxy);

	    res.x += subh_cache(h.x);
	    res.y += subh_cache(hx.x);
	    res.z += subh_cache(hy.x);
	    res.w += subh_cache(hxy.x);

	    res.x += subh_cache(h.y);
	    res.y += subh_cache(hx.y);
	    res.z += subh_cache(hy.y);
	    res.w += subh_cache(hxy.y);

	    res.x += subh_cache(h.z);
	    res.y += subh_cache(hx.z);
	    res.z += subh_cache(hy.z);
	    res.w += subh_cache(hxy.z);

	    res.x += subh_cache(h.w);
	    res.y += subh_cache(hx.w);
	    res.z += subh_cache(hy.w);
	    res.w += subh_cache(hxy.w);
	}

        barrier(CLK_LOCAL_MEM_FENCE);
    }
    
    slice[slice_width * (idy    ) + idx] = res.x;
    slice[slice_width * (idy    ) + idx + 1] = res.y;
    slice[slice_width * (idy + 1) + idx] = res.z;
    slice[slice_width * (idy + 1) + idx + 1] = res.w;
}
