/* 
 This is attempt to use __shfl to cache parts of the shared memory cache in the registers.
 We are going by half-wraps. Each half-wrap is responsible for dual-area of (4x4) * (4x4) 
 pixels (vertically  stacked). I.e. we have 32x16 block of pixels per half-wrap. The full
 wrap encloses 32x32 square. No we are using a full wrap to cache the required part of
 cache for each of the half-wraps. A bit of doubling, but may be expected to pay off.
 There is no way round: 32x32 will require 48 positions and we can't store it. The same
 way processing only half-wraps will require 24 positions, but we would have only 16
 places...
 
 Possible benefit of the approach is reduction of LD units usage and possible synergy
 with texture fetch algorithm.
 
 This code works somehow, but slow and probably has some artifacts. May be it is possible
 to optimize somehow access to shared memory... 
 
*/

__global__ static void hst_cuda_linear_kernel(int num_proj, int num_bins, float *d_SLICE, float apos_off_x, float apos_off_y, int batch) {
    float4 res = {0.f, 0.f, 0.f, 0.f};
    
    const int tidx = threadIdx.x;
    const int tidy = threadIdx.y;

    const int bidx = 2 * blockIdx.x * BLOCK_SIZE_X;
    const int bidy = batch + 2 * blockIdx.y * BLOCK_SIZE_Y;

    const int idx = bidx + 2 * tidx;
    const int idy = bidy + 2 * tidy;

    const float bx = bidx + apos_off_x;
    const float by = bidy + apos_off_y;

    const float x = idx + apos_off_x;
    const float y = idy + apos_off_y;


    __shared__  float   cache_subh[2*BLOCK_SIZE_X][BLOCK_SIZE_Y + 1];
    __shared__  float   cache_sin[BLOCK_SIZE_Y];
    __shared__  float2   cache[BLOCK_SIZE_Y][3 * BLOCK_SIZE_X + 1];
    
//#pragma unroll 8
    float projf = tidy + 0.5;
    for (int proje=0; proje<num_proj; proje += BLOCK_SIZE_Y) {
        float4 all = c_all[proje + tidy];
	int minh = (int)floor(all.z + bx * all.x - by * all.y + all.w);

	float subh = (all.z + x * all.x - minh);
	cache_subh[2*tidx][tidy] = subh;
	cache_subh[2*tidx+1][tidy] = subh + all.x;
	if (tidx == 0) cache_sin[tidy] = all.y;


#pragma unroll 3
	for (int i = 0; i < 3; i++) {
	    int pos = i * BLOCK_SIZE_X + tidx;
	    cache[tidy][pos].x = tex2D(tex_projes, minh + pos, proje + projf);
	}

	__syncthreads();

#pragma unroll 3
	for (int i = 0; i < 3; i++) {
	    int pos = i * BLOCK_SIZE_X + tidx;
	    cache[tidy][pos].y = cache[tidy][pos + 1].x - cache[tidy][pos].x;
	}

	__syncthreads();

#pragma unroll 16
	for (int i = 0; i < BLOCK_SIZE_Y; i++) {
	    float2 c;
	    float4 v1, v2;
	    float4 minh = {
		cache_subh[2 * tidx][i] - (y) * cache_sin[i],
		cache_subh[2 * tidx + 1][i] - (y) * cache_sin[i],
		cache_subh[2 * tidx][i] - (y + 1) * cache_sin[i],
		cache_subh[2 * tidx + 1][i] - (y + 1) * cache_sin[i],
	    };

	    c = cache[i][(int)minh.x];
	    v1.x = c.x; v2.x = c.y;
	    c = cache[i][(int)minh.y];
	    v1.y = c.x; v2.y = c.y;
	    c = cache[i][(int)minh.z];
	    v1.z = c.x; v2.z = c.y;
	    c = cache[i][(int)minh.w];
	    v1.w = c.x; v2.w = c.y;
	    
	    res += v1 + (minh - floorf(minh)) * v2;
	}

	__syncthreads();

    }

    *(float2*)(d_SLICE + BLOCK_SIZE_X * 2 * gridDim.x * (idy    ) + idx) = make_float2(res.x, res.y);
    *(float2*)(d_SLICE + BLOCK_SIZE_X * 2 * gridDim.x * (idy + 1) + idx) = make_float2(res.z, res.w);
}


__device__ static void hst_cuda_mplinear_kernel_2(int num_proj, int num_bins, float *d_SLICE, float apos_off_x, float apos_off_y, int batch) {
    __shared__ float f_minh[128];
    __shared__ float2 cache_offset[2][2][2][32];        //wrapy * wrapx * halfwrap * proj
    __shared__ uchar2 cache_ioffset[2][2][2][32];        //wrapy * wrapx * halfwrap * proj
    __shared__ float2 cache[IPT * PROJ_LINE];

    float res1[2 * PPT][PPT] = {0.f};
//    float res2[PPT][PPT] = {0.f};
    
    const int bidx = PPT * blockIdx.x * BLOCK_SIZE_X;
    const int bidy = batch + PPT * blockIdx.y * BLOCK_SIZE_Y;

    const float bx = bidx + apos_off_x;
    const float by = bidy + apos_off_y;

    const int tidx = threadIdx.x;
    const int tidy = threadIdx.y;

    const int halfwrap = tidy&1;
    const int wrapid = halfwrap * 16 + tidx;
    const int wrapx = (tidy/2)%2;
    const int wrapy = (tidy/2)/2;

        // block-wide minh computation
    const int ttidx = 16 * tidy + tidx;

        // texture fetches
    const int tbidx = tidy%(PPT/2);
    const int tbidy = tidy/(PPT/2);

        // sidx computation and offset caching
    const int sbidx = 4 * (tidy % 4);
    const int sbidy = 4 * (tidy / 4);

        // sidx computation only
    const int stidx = tidx % 4;
    const int stidy = tidx / 4;

        // idx computation only
    const int sidx = (sbidx + stidx);
    const int sidy = 2 * sbidy + stidy;

        // x computation and write-out
    const int idx = bidx + PPT * sidx;
    const int idy = bidy + PPT * sidy;

    const float x = idx + apos_off_x;
    const float y = idy + apos_off_y;

//    const int sbidy = 4 * ((tidy / 2) / 2);

//    const float sbx = bx + BLOCK_SIZE_X * sbidx;
//    const float sby = by + BLOCK_SIZE_Y * sbidy;
    
//    const int stidy = (2 * (tidy&1) + tidx / 8);


//    const int idx = bidx + PPT * sidx;
//    const int idy = bidy + PPT * sidy;


    const float exp23 = exp2(23.f);

    float projf = tbidy + 0.5f;
    const int num_proj_blocks = num_proj / 32;

    for (int proj_block=0; proj_block<num_proj_blocks; proj_block += 4) {
        const int proj = proj_block * 32;

        float4 all = c_all[proj + ttidx];
        float minh = floor(all.z + bx * all.x - by * all.y + all.w);
        f_minh[ttidx] = minh;

        __syncthreads();

        float fminh[4];

#pragma unroll 4
        for (int i = 0; i < 4; i++) {
            fminh[i] = f_minh[32 * i + wrapid];
        }

        int max_proj = min(num_proj_blocks - proj_block, 4);

#pragma unroll 1
        for (int subproj32 = 0; subproj32 < max_proj; subproj32++) {

                        // all.w/4 to compensate smaller size
            float4 all = c_all[proj + 32 * subproj32 + wrapid];

            float lminh = __shfl(fminh[subproj32], wrapid, 32);

            float2 offsets1 = make_float2(
                floorf(all.z + (bx + 4 * PPT * (2 * wrapx + 0)) * all.x - (by + 4 * PPT * (2 * wrapy + 0)) * all.y + (0.25f * all.w) - lminh),
                floorf(all.z + (bx + 4 * PPT * (2 * wrapx + 0)) * all.x - (by + 4 * PPT * (2 * wrapy + 1)) * all.y + (0.25f * all.w) - lminh)
            );
            float2 offsets2 = make_float2(
                floorf(all.z + (bx + 4 * PPT * (2 * wrapx + 1)) * all.x - (by + 4 * PPT * (2 * wrapy + 0)) * all.y + (0.25f * all.w) - lminh),
                floorf(all.z + (bx + 4 * PPT * (2 * wrapx + 1)) * all.x - (by + 4 * PPT * (2 * wrapy + 1)) * all.y + (0.25f * all.w) - lminh)
            );

            cache_offset[wrapy][wrapx][0][wrapid] = offsets1;
            cache_offset[wrapy][wrapx][1][wrapid] = offsets2;


            cache_ioffset[wrapy][wrapx][0][wrapid] = make_uchar2(
                (unsigned char)floor(offsets1.x),
                (unsigned char)floor(offsets1.y)
            );

            cache_ioffset[wrapy][wrapx][1][wrapid] = make_uchar2(
                (unsigned char)floor(offsets2.x),
                (unsigned char)floor(offsets2.y)
            );

            __syncthreads();

#pragma unroll 1
            for (int subproj = 0; subproj < (32/IPT); subproj++) {
                const int loop_proj = 32 * subproj32 + IPT * subproj;
                const int proje = proj + loop_proj;

                float4 all = c_all[proje + tbidy];

                int minh = __shfl(fminh[subproj32], IPT * subproj + tbidy, 32);


#pragma unroll 3
                for (int i = 0; i < 3; i++) {
	            int pos = (i * (PPT/2) + tbidx) * BLOCK_SIZE_X + tidx;
	        
	            cache[PROJ_LINE * tbidy + pos].x = tex2D(tex_projes, minh + pos, proje + projf);
                }

            
                // we may use fence instead
	        //__syncthreads();

#pragma unroll 3
                for (int i = 0; i < 3; i++) {
	            int pos = (i * (PPT/2) + tbidx) * BLOCK_SIZE_X + tidx;
                    cache[tbidy * PROJ_LINE + pos].y = cache[tbidy * PROJ_LINE + pos + 1].x - cache[tbidy * PROJ_LINE + pos].x;
                }

	        __syncthreads();

#pragma unroll 4 // IPT
	        for (int p = 0; p < IPT; p++) {
                    float4 all = c_all[proje + p];

                    float minh = __shfl(fminh[subproj32], IPT * subproj + p, 32);


                    
                        // try to make intermediate step caching in vars
                    float2 offsets_my = cache_offset[wrapy][wrapx][halfwrap][IPT * subproj];
                    float2 offsets_foreign = cache_offset[wrapy][wrapx][!halfwrap][IPT * subproj];
                    uchar2 ioffsets1 = cache_ioffset[wrapy][wrapx][0][IPT * subproj];
                    uchar2 ioffsets2 = cache_ioffset[wrapy][wrapx][1][IPT * subproj];
                    
                    float2 lcache1 = make_float2(
                        cache[p * PROJ_LINE + ioffsets1.x + wrapid].x,
                        cache[p * PROJ_LINE + ioffsets1.y + wrapid].x
                    ); 
                    float2 lcache2 = make_float2(
                        cache[p * PROJ_LINE + ioffsets2.x + wrapid].x,
                        cache[p * PROJ_LINE + ioffsets2.y + wrapid].x
                    );
                    
                    
                    
                    float h = all.z + x * all.x - y * all.y - minh;

#pragma unroll 4 // PPT
	            for (int i = 0; i < PPT; i++) {
#pragma unroll 4 // PPT
	                for (int j = 0; j < PPT; j++) {
	                    float subh1 = h - (i          ) * all.y + j * all.x - offsets_my.x;
	                    float subh2 = h - (i + 4 * PPT) * all.y + j * all.x - offsets_my.y;

                            float fsubh1 = subh1 + exp23;
                            int idx1 = (*(int*)(&fsubh1)) - 0x4B000000;
                            fsubh1 = subh1 - (fsubh1 - exp23);

                            float fsubh2 = subh2 + exp23;
                            int idx2 = (*(int*)(&fsubh2)) - 0x4B000000;
                            fsubh2 = subh2 - (fsubh2 - exp23);

//                            int foreign1 = __shfl_xor(idx1, 0x10, 32);
//                            int foreign2 = __shfl_xor(idx2, 0x10, 32);

                            float2 c, c1, c2;

                            c1.x = __shfl(lcache1.x, idx1, 32);
                            c2.x = __shfl(lcache2.x, idx1, 32);
                            
                            c1.y = __shfl(lcache1.y, idx2, 32);
                            c2.y = __shfl(lcache2.y, idx2, 32);
                            
                            //c = halfwrap?c2:c1;
                            
                            res1[i][j] += (halfwrap)?c2.x:c1.x;
                            res1[4 + i][j] += (halfwrap)?c2.y:c1.y;


/*
	                    float2 c1 = cache[p * PROJ_LINE + idx1];
	                    res1[i][j] += c1.x + fsubh1 * c1.y;

	                    float2 c2 = cache[p * PROJ_LINE + idx2];
	                    res1[4+i][j] += c2.x + fsubh2 * c2.y;
*/
	                }
	            }
	        }
	    }


	    __syncthreads();
	}
    }


    float4 *result = (float4*)(d_SLICE + PPT * BLOCK_SIZE_X * gridDim.x * idy + idx);
#pragma unroll 4
    for (int i = 0; i < PPT; i++) {
	result[i * gridDim.x * BLOCK_SIZE_X] = *(float4*)&res1[i][0];
	result[(i + 16) * gridDim.x * BLOCK_SIZE_X] = *(float4*)&res1[4 + i][0];
    }
}


__global__ static void hst_cuda_mplinear_kernel(int num_proj, int num_bins, float *d_SLICE, float apos_off_x, float apos_off_y, int batch) {
    const int mode = 0;//((gridDim.x * blockIdx.y + blockIdx.x)/8) % 2;

    if (mode) {
        hst_cuda_mplinear_kernel_1(num_proj, num_bins, d_SLICE, apos_off_x, apos_off_y, batch);
    } else {
        hst_cuda_mplinear_kernel_2(num_proj, num_bins, d_SLICE, apos_off_x, apos_off_y, batch);
    }
}

