
#include"Python.h"
#include"structmember.h"
#include<string.h>

#include "numpy/arrayobject.h"

#ifdef LINUX
 #define _FILE_OFFSET_BITS  64
#endif


#define DEBUG(a)


 /* The error object to expose 
 */

static PyObject *ErrorObject;
#define onError(message)\
  { PyErr_SetString(ErrorObject, message); return NULL;}


#define  __CCD_Filter_name__  "CCD_Filter2"
#define __init__Filter__      initCCD_Filter2

static char Filter_doc[]=
"/* "
""
"   Filter( item , pos_edf_, size_edf_, pos_edf, size_edf,   CCD_FILTER_PARA     )"
"  "
"   CCD_FILTER_PARA  is a python dictionary with parameters"
""
"  VERSION 2"
" */"
;

static PyObject *
Filter(PyObject *self_a, PyObject *args)
{
  PyObject * A         ;
  PyObject * pos_      ;
  PyObject * siz_      ;
  PyObject * pos       ;
  PyObject * siz       ;
  PyObject * para_dict ;
  PyObject * tmp;
  char msg[400];
  
  int x0_,x1_, x0,x1, s0_,s1_,s0,s1;
  int X0,X1,X0_, X1_;
  int dim0, dim1;
  int nsum;
  double average;


  int c0,c1;
  int c0_,c1_;




  PyObject * dic_key;
  PyObject * dic_value;
  double threshold ;

  int i;


  int *iptr[] = { &x0_,&x1_, &x0,&x1, &s0_,&s1_,&s0,&s1  };

  int shift_x[] = { -1,-1, 1, 1  };
  int shift_y[] = { -1,+1,-1,+1  };
  int si;

  PyObject ** containers[] = {&pos_ , &pos, &siz_, &siz   };


  float * a_data;





 if(!PyArg_ParseTuple(args,"OOOOOO:CDD_Filter",&A,&pos_,&siz_, &pos, &siz, &para_dict  ))
   return NULL;

 /* check the Objects */

  if(!PyArray_Check(A ))    onError("not a PyArray, argument 1");
  if(!PyList_Check (pos_ ) )     onError("not a PyList, argument 2");
  if(!PyList_Check (siz_ ))    onError("not a PyList, argument 3");
  if(!PyList_Check (pos ))         onError("not a PyList, argument 4");
  if(!PyList_Check(siz ))       onError("not a PyList, argument 5");
  if(!PyDict_Check(para_dict ))        onError("not a PyDict, argument 6");

  /* check the types */

  if( ((PyArrayObject *) A )  ->descr->type_num != PyArray_FLOAT ) onError(" arg 1 is not an array of float " ) ;

  /* check n of dimensione */

  if( ((PyArrayObject *) A)->nd !=2)  onError(" arg 1 is not a matrix " ) ;
 

  /* check contiguity */

  if(  ((PyArrayObject *) A) ->flags %2 == 0) onError(" array has to be contiguous");

  a_data = (float *) ((PyArrayObject *) A)->data;
 
  /* getting positions and sizes */

  for (i=0; i<8; i++) {
    tmp =  PyList_GetItem( *(containers[i/2])  , i%2);
    if(!PyInt_Check( tmp )) {
      sprintf(msg, " not an integer, item %d of argument %d", i%2, i/2+2);
      onError(msg);
    }
    *iptr[i] = PyInt_AS_LONG( tmp ); 
  }


  /* getting parameter(s) contained in the dictionary */

  dic_key = PyString_FromString("threshold");
  dic_value =  PyDict_GetItem(para_dict , dic_key);
  Py_DECREF(dic_key);

  if( dic_value == NULL) 
    onError(" unable to find the threshold key in the passed dictionary");

  if( !PyFloat_Check(dic_value) ) 
    onError("supplied value for threshold is not a float ");
  
  threshold = PyFloat_AsDouble (dic_value  );

  

  /* setting better the limits */

  x0 = x0 - x0_ ;
  x1 = x1 - x1_ ;
  
  x0_ = 0 ;
  x1_ = 0 ;
  

  X0 = x0 + s0 ;
  X1 = x1 + s1 ;
  
  X0_ = x0_ + s0_ ;
  X1_ = x1_ + s1_ ;


  /* checking that the supllied limits are not completely wrong */

  if ( ((PyArrayObject *) A)->dimensions[0] != s0_  || ((PyArrayObject *) A)->dimensions[1] != s1_  ) {
    onError(" array dimension and supplied limits are not consistent");
  }
  
  dim0 = s0_;
  dim1 = s1_;

  /* straighforward addressing scheme in case of contiguous memory. Or you'll need to use strides */
#define AA(i,j)  a_data[i*dim1+j    ]


  for(c0=x0+1 ; c0< X0 -1; c0++) {
    
    for(c1=x1+1; c1< X1 -1; c1++) {
      
      average=( AA(c0+1,c1+1 )+ AA(c0+1,c1-1 )+AA(c0-1,c1+1 )+AA(c0-1,c1-1 ) )/4.0;
      if (  AA(c0,c1 ) < average*threshold ) {
	AA(c0,c1 )= average;
      }
    } 
  }
  
  for(c0= x0 ; c0<X0 ; c0 +=(s0-1)) {  
    for(c1=x1; c1<X1; c1 +=(s1-1) ) {   
      nsum=0;
      average=0.0;
      for(si=0; si<4; si++) {
	c0_ = c0+ shift_x[si];
	c1_ = c1+ shift_y[si];
	if( c0_>=x0_ &&  X0_ <s0_ &&    c1_>=x1 &&  c1_ <X1_  ) {
	  average += AA(c0_,c1_) ;
	  nsum++;
	}
      }
      average/=nsum;
      if (  AA(c0,c1 ) < average*threshold ) {
	AA(c0,c1 )= average;
      }
    }
  } 
  
  
  Py_INCREF(Py_None);
  return Py_None;


}



static PyMethodDef CCD_Filter_functions [] = {
  {"Filter", Filter,  METH_VARARGS, Filter_doc},
  { NULL, NULL}
};






void 
__init__Filter__   ()
{
  PyObject *m, *d;
  m = Py_InitModule(  __CCD_Filter_name__    ,  CCD_Filter_functions  );
  d = PyModule_GetDict(m);
  ErrorObject = Py_BuildValue("s",  __CCD_Filter_name__  ".error");
  PyDict_SetItemString(d,"error", ErrorObject);
  if(PyErr_Occurred())
    Py_FatalError("can't initialize module " __CCD_Filter_name__  );

#ifdef import_array
  import_array();
#endif
}

