#define HST_CUDA_ARRAY				//!< Use cudaArray instead of binding textures
#define HST_TEX_KERNEL				//!< Enforce ppt1 tex-based kernel
//#define HST_BASE_KERNEL				//!< Enforce simplest kernel

//#define HST_OPTIMIZE_KEPLER

#if SLICE_BLOCK > 1
# define HST_CUDA_ARRAY
# define HST_TEX_KERNEL
#endif /* SLICE_BLOCK */

#ifdef HST_OPTIMIZE_KEPLER
# define HST_CREATE_TEXTURE			//!< Create texture instead of binding. Seems faster, but only supported since Kepler
# define HST_SHFL_SUM				//!< Important as ShMem is shared with L1 (in case of spilage)
# define HST_SET_KEPLER_BOUNDS
# if (defined(HST_SET_KEPLER_BOUNDS)&&(SLICE_BLOCK > 1))
#  define HST_KEPLER_PREFER_L1
# endif
#endif

#ifdef HST_BASE_KERNEL
# define HST_TEX_KERNEL
#endif



#define MAXNPROJECTIONS 4096			//!< Maximum number of supported projections per slices
#define HST_CUDA_MAX_DEVICES 16			//!< Maximum CUDA-devices supported by the module (check tex_projes)

#define BLOCK_SIZE 16				//!< CUDA block size (for filtering)
#define BLOCK_SIZE_X BLOCK_SIZE			//!< Check kernels, there are some expectations
#define BLOCK_SIZE_Y BLOCK_SIZE			//!< Check kernels, there are some expectations

#define CUDA_HEATUP_MAX_CHANGE 0.05
#define CUDA_HEATUP_MIN_ITERATIONS 200
#define CUDA_HEATUP_MAX_ITERATIONS 250
#define CUDA_HEATUP_KERNEL_ITERATIONS 4096
#define CUDA_HEATUP_ELEMENTS 4
#define CUDA_HEATUP_GRID 4480 // 16 * 5 * 7 (covers SM numbers) * 8 (covers warps per SM)
#define CUDA_MAX_STATS 8000

    /* We should be carefoul here, in many cases it faster to preform everything at once
    probably due to the texture caching */
//#ifdef HST_OPTIMIZE_KEPLER
//# define BP_BATCH_SIZE 128			//!< Batch size (in blocks) for Back Projection  (should be even)
//#else /* HW_OPTIMIZE_KEPLER */
//# ifdef  HW_IGNORE_OLD_HARDWARE
//#  define BP_BATCH_SIZE 128			//!< Batch size (in blocks) for Back Projection  (should be even)
//# else /*  HW_IGNORE_OLD_HARDWARE */
//#  define BP_BATCH_SIZE 8			//!< Batch size (in blocks) for Back Projection  (should be even)
//# endif /*  HW_IGNORE_OLD_HARDWARE */
//#endif /*  HW_OPTIMIZE_KEPLER */
//#define FFT_BATCH_SIZE 8			//!< Batch size (in blocks) for FFT transformations (should be even)


#define BP_BATCH_SIZE 512
#define FFT_BATCH_SIZE 256                      //!< Anyway limited by size of constant memory
