__global__ static void  hst_cuda_unpack_to_half_kernel(half *out, int dpitch, cufftReal *data, int spitch) {
    float tmp1, tmp2;

    int  idx = blockIdx.x * BLOCK_SIZE + threadIdx.x;
    int  idy = blockIdx.y * BLOCK_SIZE + threadIdx.y;

    int pos = 2 * (idy * spitch + idx);

    tmp1 = data[pos];
    tmp2 = data[pos + 1];

    pos = 2 * idy * dpitch + idx ;

    out[pos] = __float2half(tmp1);
    out[pos + dpitch] = __float2half(tmp2);
}

#define PPT 4
// Iterations per line
#define IPT (16/PPT)
//#define PROJ_LINE (2 * IPT * PPT * 3)
#define PROJ_LINE (PPT * BLOCK_SIZE_X * 3 / 2)


__global__ static void hst_cuda_mplinear_half_kernel(int num_proj, int num_bins, float *d_SLICE, float apos_off_x, float apos_off_y, int batch) {
    half2 res[PPT][PPT] = {0.f};
/*
    half res1[PPT][PPT] = {0.f};
    half res2[PPT][PPT] = {0.f};
*/

    const int tidx = threadIdx.x;
    const int tidy = threadIdx.y;

    const int bidx = PPT * blockIdx.x * BLOCK_SIZE_X;
    const int bidy = batch + PPT * blockIdx.y * BLOCK_SIZE_Y;


    const float bx = bidx + apos_off_x;
    const float by = bidy + apos_off_y;

    __shared__ float f_minh[128];
    __shared__ half   cache1[IPT * PROJ_LINE + 1];
    __shared__ half   cache2[IPT * PROJ_LINE + 1];
    
    const int tbidy = tidy/(PPT/2);
    const int tbidx = tidy%(PPT/2);
    
    const int wrapid = (tidy&1) * 16 + tidx;
    
    const int sbidx = 8 * ((tidy / 2) % 2);
    const int sbidy = 8 * ((tidy / 2) / 2);
    
    const int stidx = tidx % 8;
    const int stidy = 2 * (2 * (tidy&1) + tidx / 8);

    const int sidx = (sbidx + stidx);
    const int sidy = (sbidy + stidy);

    const int idx = bidx + sidx;
    const int idy = bidy + sidy;

    const float x = idx + apos_off_x;
    const float y = idy + apos_off_y;

    const float exp23 = exp2(23.f);
    
    const int ttidx = 16 * tidy + tidx;

    float projf = tbidy + 0.5f;
    
    const int num_proj_blocks = num_proj / 32;
    
    for (int proj_block=0; proj_block<num_proj_blocks; proj_block += 4) {
        const int proj = proj_block * 32;
        
        float4 all = c_all[proj + ttidx];
        float minh = floor(all.z + bx * all.x - by * all.y + all.w);
        f_minh[ttidx] = minh;
        
        __syncthreads();

        float fminh[4];

#pragma unroll 4
        for (int i = 0; i < 4; i++) {
            fminh[i] = f_minh[32 * i + wrapid];
        }

        int max_proj = min(num_proj_blocks - proj_block, 4);

#pragma unroll 1
        for (int subproj32 = 0; subproj32 < max_proj; subproj32++) {
#pragma unroll 1
            for (int subproj = 0; subproj < (32/IPT); subproj++) {
                const int loop_proj = 32 * subproj32 + IPT * subproj;
                const int proje = proj + loop_proj;

                float4 all = c_all[proje + tbidy];

	        int minh = __shfl(fminh[subproj32], IPT * subproj + tbidy, 32);

#pragma unroll 3
                for (int i = 0; i < 3; i++) {
	            int pos = (i * (PPT/2) + tbidx) * BLOCK_SIZE_X + tidx;
	            cache1[PROJ_LINE * tbidy + pos] = __float2half(tex2D(tex_projes, minh + pos, proje + projf));
                }
            
            // we may use fence instead
	    //__syncthreads();

#pragma unroll 3
                for (int i = 0; i < 3; i++) {
	            int pos = (i * (PPT/2) + tbidx) * BLOCK_SIZE_X + tidx;
                    cache2[tbidy * PROJ_LINE + pos] = __hsub(cache1[tbidy * PROJ_LINE + pos + 1], cache1[tbidy * PROJ_LINE + pos]);
                }

	        __syncthreads();

#pragma unroll 4 // IPT
	        for (int p = 0; p < IPT; p++) {
                    float4 all = c_all[proje + p];

                    float minh = __shfl(fminh[subproj32], IPT * subproj + p, 32);


                    float h = all.z + x * all.x - y * all.y - minh;


//                    char *cache_row = ((char*)(&cache)) + p * PROJ_LINE * sizeof(float2);


#pragma unroll 4 // PPT
	            for (int i = 0; i < PPT; i++) {
	                float subh = h;
	                h -= 16.f * all.y;
#pragma unroll 4 // PPT
	                for (int j = 0; j < PPT; j++) {
	                    float subh1 = subh;
	                    float subh2 = subh - all.y;


                            float fsubh1 = subh1 + exp23;
                            int idx1 = (*(int*)(&fsubh1)) - 0x4B000000;
                            fsubh1 = subh1 - (fsubh1 - exp23);

			    float fsubh2 = subh2 + exp23;
                            int idx2 = (*(int*)(&fsubh2)) - 0x4B000000;
                            fsubh2 = subh2 - (fsubh2 - exp23);

			    half2 fsubh = __floats2half2_rn(fsubh1, fsubh2);
	                    half2 cx = __halves2half2(cache1[p * PROJ_LINE + idx1], cache1[p * PROJ_LINE + idx2]);
	                    half2 cy = __halves2half2(cache2[p * PROJ_LINE + idx1], cache2[p * PROJ_LINE + idx2]);
	                    res[i][j] = __hadd2(res[i][j], __hfma2(fsubh, cy, cx));

/*
	                    half c1x = cache1[p * PROJ_LINE + idx1];
	                    half c1y = cache2[p * PROJ_LINE + idx1];
	                    res1[i][j] = __hadd(res1[i][j], __hfma(__float2half(fsubh1), c1y, c1x));

	                    half c2x = cache1[p * PROJ_LINE + idx2];
	                    half c2y = cache2[p * PROJ_LINE + idx2];
	                    res2[i][j] = __hadd(res2[i][j], __hfma(__float2half(fsubh2), c2y, c2x));
*/

	                    subh += 16.f * all.x;
	                }
	            }

	        }
	    }
	    __syncthreads();
	}
    }

#pragma unroll 4
    for (int i = 0; i < PPT; i++) {
#pragma unroll 4
	for (int j = 0; j < PPT; j++) {
	        // or vice-versa
	    d_SLICE[BLOCK_SIZE_X * PPT * gridDim.x * (idy + i * BLOCK_SIZE_Y) + idx + j * BLOCK_SIZE_X] = __low2float(res[i][j]);
	    d_SLICE[BLOCK_SIZE_X * PPT * gridDim.x * (idy + i * BLOCK_SIZE_Y + 1) + idx + j * BLOCK_SIZE_X] = __high2float(res[i][j]);
/*
	    d_SLICE[BLOCK_SIZE_X * PPT * gridDim.x * (idy + i * BLOCK_SIZE_Y) + idx + j * BLOCK_SIZE_X] = __half2float(res1[i][j]);
	    d_SLICE[BLOCK_SIZE_X * PPT * gridDim.x * (idy + i * BLOCK_SIZE_Y + 1) + idx + j * BLOCK_SIZE_X] = __half2float(res2[i][j]);
*/
	}
    }
}


/*
This is unfinished and broken
__global__ static void hst_cuda_mplinear_fullhalf_kernel(int num_proj, int num_bins, float *d_SLICE, float apos_off_x, float apos_off_y, int batch) {
    half2 res[PPT][PPT] = {0.f};

    const int tidx = threadIdx.x;
    const int tidy = threadIdx.y;

    const int bidx = PPT * blockIdx.x * BLOCK_SIZE_X;
    const int bidy = batch + PPT * blockIdx.y * BLOCK_SIZE_Y;


    const half bx = __hadd(__int2half_rn(bidx), __float2half(apos_off_x));
    const half by = __hadd(__int2half_rn(bidy), __float2half(apos_off_y));

    __shared__ half f_minh[128];
    __shared__ half cache1[IPT * PROJ_LINE + 1];
    __shared__ half cache2[IPT * PROJ_LINE + 1];
    
    const int tbidy = tidy/(PPT/2);
    const int tbidx = tidy%(PPT/2);
    
    const int wrapid = (tidy&1) * 16 + tidx;
    
    const int sbidx = 8 * ((tidy / 2) % 2);
    const int sbidy = 8 * ((tidy / 2) / 2);
    
    const int stidx = tidx % 8;
    const int stidy = 2 * (2 * (tidy&1) + tidx / 8);

    const int sidx = (sbidx + stidx);
    const int sidy = (sbidy + stidy);

    const int idx = bidx + sidx;
    const int idy = bidy + sidy;

    const half x = __hadd(__int2half_rn(idx), __float2half(apos_off_x));
    const half y = __hadd(__int2half_rn(idy), __float2half(apos_off_y));

    const float exp23 = exp2(23.f);
    const half2 exp10 = __floats2half2_rn(exp2(10.f), exp2(10.f));
    
    const int ttidx = 16 * tidy + tidx;

    float projf = tbidy + 0.5f;
    
    const int num_proj_blocks = num_proj / 32;
    
    for (int proj_block=0; proj_block<num_proj_blocks; proj_block += 4) {
        const int proj = proj_block * 32;
        
        half minh = hfloor(__hfma(bx, all(0), __hfma(__hneg(by), all(1), __hadd(all(2), all(3))))); //   all.z + bx * all.x - by * all.y + all.w);
        f_minh[ttidx] = minh;
        
        __syncthreads();

        half fminh[4];

#pragma unroll 4
        for (int i = 0; i < 4; i++) {
            fminh[i] = f_minh[32 * i + wrapid];
        }

        int max_proj = min(num_proj_blocks - proj_block, 4);

#pragma unroll 1
        for (int subproj32 = 0; subproj32 < max_proj; subproj32++) {
#pragma unroll 1
            for (int subproj = 0; subproj < (32/IPT); subproj++) {
                const int loop_proj = 32 * subproj32 + IPT * subproj;
                const int proje = proj + loop_proj;

	        half minh = __shfl(fminh[subproj32], IPT * subproj + tbidy, 32);

#pragma unroll 3
                for (int i = 0; i < 3; i++) {
	            int pos = (i * (PPT/2) + tbidx) * BLOCK_SIZE_X + tidx;
		    cache1[PROJ_LINE * tbidy + pos] = __float2half(tex2D(tex_projes, __half2int_rn(minh) + pos, proje + projf));
                }
            
            // we may use fence instead
	    //__syncthreads();

#pragma unroll 3
                for (int i = 0; i < 3; i++) {
	            int pos = (i * (PPT/2) + tbidx) * BLOCK_SIZE_X + tidx;
                    cache2[tbidy * PROJ_LINE + pos] = __hsub(cache1[tbidy * PROJ_LINE + pos + 1], cache1[tbidy * PROJ_LINE + pos]);
                }

	        __syncthreads();

#pragma unroll 4 // IPT
	        for (int p = 0; p < IPT; p++) {
//                    half all[4] = c_all[proje + p];

                    half minh = __shfl(fminh[subproj32], IPT * subproj + p, 32);

                    half h = hfloor(__hfma(x, pall(0), __hneg(__hfma(y, pall(1), __hsub(minh, pall(2)))))); //   all.z + x * all.x - y * all.y - minh);
                    half ally16 = __hmul(__float2half(16.f), pall(1));
                    half allx16 = __hmul(__float2half(16.f), pall(0));

//                    half2 half_rounding = __halves2half2(0x25800, 0x25800);


//                    char *cache_row = ((char*)(&cache)) + p * PROJ_LINE * sizeof(float2);


#pragma unroll 4 // PPT
	            for (int i = 0; i < PPT; i++) {
	                half subh = h;
	                h = __hsub(h, ally16);
#pragma unroll 4 // PPT
	                for (int j = 0; j < PPT; j++) {
//	                    float subh1 = subh;
//	                    float subh2 = subh - all.y;
	                    half2 subh2 = __halves2half2(subh, __hsub(subh, pall(1)));
	                    half2 fsubh = __hadd2(subh2, exp10);
	                    ushort2 idx = (*(ushort2*)(&fsubh));
	                    idx.x -= 0x25800;
	                    idx.y -= 0x25800;

	                    half2 cx = __halves2half2(cache1[p * PROJ_LINE + idx.x], cache1[p * PROJ_LINE + idx.y]);
	                    half2 cy = __halves2half2(cache2[p * PROJ_LINE + idx.x], cache2[p * PROJ_LINE + idx.y]);
	                    res[i][j] = __hadd2(res[i][j], __hfma2(fsubh, cy, cx));

	                    subh = __hadd(subh, allx16);
	                }
	            }

	        }
	    }
	    __syncthreads();
	}
    }

#pragma unroll 4
    for (int i = 0; i < PPT; i++) {
#pragma unroll 4
	for (int j = 0; j < PPT; j++) {
	    d_SLICE[BLOCK_SIZE_X * PPT * gridDim.x * (idy + i * BLOCK_SIZE_Y) + idx + j * BLOCK_SIZE_X] = __low2float(res[i][j]);
	    d_SLICE[BLOCK_SIZE_X * PPT * gridDim.x * (idy + i * BLOCK_SIZE_Y + 1) + idx + j * BLOCK_SIZE_X] = __high2float(res[i][j]);
	}
    }
}

*/