/*
 * The PyHST program is Copyright (C) 2002-2011 of the
 * European Synchrotron Radiation Facility (ESRF) and
 * Karlsruhe Institute of Technology (KIT).
 *
 * PyHST is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * hst is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

static const char *hst_opencl_kernels_source = 
#include "hst_opencl_kernels.cl.h"

#define HST_MULTI_POINT_RECONSTRUCTION
#define HST_MAX_KERNEL_VARIANTS 5


struct HSTKernelDescriptionT {
    const char *kernel;
    size_t shared_memory;
};
typedef struct HSTKernelDescriptionT HSTKernelDescription;

static const HSTKernelDescription HST_KERNEL_DESCRIPTION[] = {
    {"hst_backproject_gpu_oversample4", (7 * BLOCK_SIZE * BLOCK_SIZE + 2 * BLOCK_SIZE) * sizeof(float)},
    {"hst_backproject_gpu_linear", (8 * BLOCK_SIZE * BLOCK_SIZE + 3 * BLOCK_SIZE) * sizeof(float)},
    {"hst_backproject_gpu_nearest", (5 * BLOCK_SIZE * BLOCK_SIZE + 3 * BLOCK_SIZE) * sizeof(float)},
    {"hst_backproject_fermi_oversample4", (14 * BLOCK_SIZE * BLOCK_SIZE + 3 * BLOCK_SIZE) * sizeof(float)},
    {"hst_backproject_fermi_linear", (8 * BLOCK_SIZE * BLOCK_SIZE + 5 * BLOCK_SIZE) * sizeof(float)},
    {"hst_backproject_fermi_nearest", (5 * BLOCK_SIZE * BLOCK_SIZE + 3 * BLOCK_SIZE) * sizeof(float)},
    {"hst_backproject_direct_nearest", (3 * BLOCK_SIZE + 2) * BLOCK_SIZE * sizeof(float)},
    {"hst_backproject_amd_direct_nearest", (3 * BLOCK_SIZE + 2) * BLOCK_SIZE * sizeof(float)},
    {"hst_backproject_direct_oversample4", (12 * BLOCK_SIZE + 1) * BLOCK_SIZE * sizeof(float)},
    {"hst_backproject_amd_direct_oversample4", (12 * BLOCK_SIZE + 2) * BLOCK_SIZE * sizeof(float)},
    {"hst_backproject_dma_oversample4", (12 * BLOCK_SIZE + 1) * BLOCK_SIZE * sizeof(float)},
    {"hst_backproject_direct_linear", (6 * BLOCK_SIZE + 2) * BLOCK_SIZE * sizeof(float)},
    {"hst_backproject_amd_direct_linear", (6 * BLOCK_SIZE + 2) * BLOCK_SIZE * sizeof(float)},
//    {"hst_backproject_6ppt_oversample4", 5 * BLOCK_SIZE * BLOCK_SIZE * sizeof(float)},
    {"hst_backproject_6ppt_oversample4", 5 * 8 * 8 * sizeof(float)},
    {"hst_backproject_8ppt_oversample4", 6 * 8 * 8 * sizeof(float)},
    {"hst_backproject_8ppt_linear", 2 * (2 * 8 * 8 + 1) * sizeof(float)},
/*    {"hst_backproject_8ppt_vliw_oversample4", 6 * 8 * 8 * sizeof(float)},
    {"hst_backproject_8ppt_vliw_linear", 2 * (2 * 8 * 8 + 1) * sizeof(float)},*/
    {"hst_backproject_kepler_tex", (2 * (16 * 18) + 56) * sizeof(float)},
    { NULL, 0}
};


/* NOTE: the following kernel names MUST match the preceeding IDs */
static const char *HST_KERNEL_NAMES[HST_DEVICE_NUMBER][HST_KERNEL_NUMBER][HST_MAX_KERNEL_VARIANTS] = {
    { // Default
	{"hst_filter"},
	{"hst_backproject_cpu"},
	{"hst_spread"},
	{"hst_pack"},
	{"hst_normalize_idft"}
    },{ // GPU/Default
	{},
	{"hst_backproject_gpu"},
	{},
	{},
	{}
    },{ // GT200
	{},
#ifdef HST_OPENCL_CONST_HACK
	{"hst_backproject_gpu_const"},
#else /* HST_OPENCL_CONST_HACK */
	{},
#endif /* HST_OPENCL_CONST_HACK */
	{},
	{},
	{}
    },{ // FERMI (Fermi kernels are expected to process 4 pixels per thread)
	{},
#ifdef HST_MULTI_POINT_RECONSTRUCTION
	{"hst_backproject_fermi_linear", "hst_backproject_fermi_nearest", "hst_backproject_fermi_oversample4"},
//	{"hst_backproject_direct_linear", "hst_backproject_direct_nearest", "hst_backproject_direct_oversample4"},
#else /* HST_MULTI_POINT_RECONSTRUCTION */
	{"hst_backproject_gpu_linear", "hst_backproject_gpu_nearest", "hst_backproject_gpu_oversample4"},
#endif /* HST_MULTI_POINT_RECONSTRUCTION */
	{},
	{},
	{}
    },{ // Kepler
	{},
#ifdef HST_MULTI_POINT_RECONSTRUCTION
	{"hst_backproject_6ppt_tex", "hst_backproject_fermi_nearest", "hst_backproject_6ppt_oversample4"},
#else /* HST_MULTI_POINT_RECONSTRUCTION */
	{"hst_backproject_gpu_linear", "hst_backproject_gpu_nearest", "hst_backproject_gpu_oversample4"},
#endif /* HST_MULTI_POINT_RECONSTRUCTION */
	{},
	{},
	{}
    },{ // Pascal
	{},
#ifdef HST_MULTI_POINT_RECONSTRUCTION
	{"hst_backproject_6ppt_tex", "hst_backproject_fermi_nearest", "hst_backproject_6ppt_oversample4"},
#else /* HST_MULTI_POINT_RECONSTRUCTION */
	{"hst_backproject_gpu_linear", "hst_backproject_gpu_nearest", "hst_backproject_gpu_oversample4"},
#endif /* HST_MULTI_POINT_RECONSTRUCTION */
	{},
	{},
	{}
    },{ // Cypress
	{},
#ifdef HST_MULTI_POINT_RECONSTRUCTION
	{"hst_backproject_4ppt_tex", "hst_backproject_amd_direct_nearest", "hst_backproject_direct_oversample4"},
//	{"hst_backproject_4ppt_tex", "hst_backproject_amd_direct_nearest", "hst_backproject_8ppt_oversample4"},
//	{"hst_backproject_6ppt_vliw_tex", "hst_backproject_amd_direct_nearest", "hst_backproject_8ppt_vliw_oversample4"},
#else /* HST_MULTI_POINT_RECONSTRUCTION */
	{"hst_backproject_kepler_tex", "hst_backproject_gpu_nearest", "hst_backproject_gpu_oversample4"},
//	{"hst_backproject_gpu", "hst_backproject_gpu_nearest", "hst_backproject_gpu_oversample4"},
#endif /* HST_MULTI_POINT_RECONSTRUCTION */
	{},
	{},
	{}
    },{ // GCN
	{},
#ifdef HST_MULTI_POINT_RECONSTRUCTION
//	{"hst_backproject_amd_direct_linear", "hst_backproject_amd_direct_nearest", "hst_backproject_amd_direct_oversample4"},
	{"hst_backproject_direct_linear", "hst_backproject_amd_direct_nearest", "hst_backproject_direct_oversample4"},
#else /* HST_MULTI_POINT_RECONSTRUCTION */
	{"hst_backproject_gpu", "hst_backproject_gpu_nearest", "hst_backproject_gpu_oversample4"},
//	{"hst_backproject_gpu_linear", "hst_backproject_gpu_nearest", "hst_backproject_gpu_oversample4"},
//	{"hst_backproject_kepler_tex", "hst_backproject_gpu_nearest", "hst_backproject_gpu_oversample4"},
#endif /* HST_MULTI_POINT_RECONSTRUCTION */
	{},
	{},
	{}
    },{ // AMD/CPU
	{},
	{},
	{},
	{},
	{}
    },{ // Intel/CPU
	{},
	{},
	{},
	{},
	{}
    }
};

inline const HSTKernelDescription *hst_opencl_resolve_kernel_description(OpenCLContext *ctx, const char *name) {
    int i;
    
    for (i = 0; HST_KERNEL_DESCRIPTION[i].kernel; i++) {
	if (!strcmp(HST_KERNEL_DESCRIPTION[i].kernel, name)) return HST_KERNEL_DESCRIPTION + i;
    }
    
    return NULL;
}


inline hst_device_type hst_opencl_resolve_kernel_architecture(OpenCLContext *ctx, hst_kernel_type kernel_id) {
    if (HST_KERNEL_NAMES[ctx->arch][kernel_id][0]) return ctx->arch;
    if ((ctx->device_type == CL_DEVICE_TYPE_GPU)&&(HST_KERNEL_NAMES[ctx->arch][kernel_id][1])) return HST_DEVICE_GPU;
    return HST_DEVICE_UNKNOWN;
}

static const char *hst_opencl_resolve_kernel(OpenCLContext *ctx, hst_kernel_type kernel_id, int variant) {
    int i;
    hst_device_type arch = hst_opencl_resolve_kernel_architecture(ctx, kernel_id);
    if (variant >= HST_MAX_KERNEL_VARIANTS) variant = HST_MAX_KERNEL_VARIANTS - 1;
    
    for (i = 0; i < variant; i++) {
	if (!HST_KERNEL_NAMES[arch][kernel_id][i+1]) break;
    }

    return HST_KERNEL_NAMES[arch][kernel_id][i];
}

static char* hst_opencl_load_program(const char* filename)
{
    FILE *fp = fopen(filename, "r");
    if (fp == NULL) {
        pyhst_error("Error: Couldn't open %s\n", filename);
        return NULL;
    }
    fseek (fp , 0 , SEEK_END);
    const size_t len = ftell(fp);
    rewind(fp);
    char *buffer;
    buffer = (char*) malloc (len * sizeof(char));
    if (buffer == NULL)
        return NULL;
    size_t buflen = fread(buffer, 1, len, fp);
    fclose(fp);
    if (buflen != len) {
        pyhst_error("Error: Couldn't read program %s from disk\n", filename);
        free(buffer);
        return NULL;
    }
    return buffer;
}


static int hst_opencl_load_kernels(OpenCLContext *ctx, int *variants) {
    int i;
    int err;
    hst_kernel_type kernel_id;
    
    HSTSetup *setup = ((HSTReconstructorContext*)ctx)->setup;
    int num_proj = setup->num_projections;

#ifdef HST_OPENCL_LOAD_PTX
    FILE *f = fopen(HST_OPENCL_LOAD_PTX, "rb");
    if (f) {
	fseek(f, 0, SEEK_END); 
	size_t len = ftell(f);
	fseek(f, 0, SEEK_SET); 
    
	char *buf = (char*)malloc(len*sizeof(char));
	if ((buf)&&(fread(buf, 1, len, f) == len)) {
	    ctx->program = clCreateProgramWithBinary(ctx->clctx, 1, &ctx->device_id,  &len, (const unsigned char**)&buf, NULL, &err);
	    CL_CHECK_ERROR(err);
	} else {
	    pyhst_error("Error reading the PTX file %s", HST_OPENCL_LOAD_PTX);
	    return 255;
	}
	if (buf) free(buf);
	fclose(f);
    } else {
	pyhst_error("Error reading the PTX file %s", HST_OPENCL_LOAD_PTX);
	return 255;
    }
#else /* HST_OPENCL_LOAD_PTX */    
    const char* source = hst_opencl_kernels_source;

# ifdef HST_OPENCL_CONST_HACK
    size_t source_ptr = 0;
    size_t source_len = strlen(source);
    const size_t source_size = source_len + 3 * ( 11 * num_proj + 64);

    char *source_buf = malloc(source_size);
    check_alloc(source_buf, "source_buf");

    sprintf(source_buf + source_ptr, "__constant unsigned cos_itable[] = {0x%08x", ((uint32_t*)setup->cos_s)[0]); source_ptr += strlen(source_buf + source_ptr);
    for (i = 1; i < num_proj; i++) {
	sprintf(source_buf + source_ptr, ",0x%08x", ((uint32_t*)setup->cos_s)[i]); source_ptr += 11;
    }
    sprintf(source_buf + source_ptr, "};\n"); source_ptr += strlen(source_buf + source_ptr);

    sprintf(source_buf + source_ptr, "__constant unsigned sin_itable[] = {0x%08x", ((uint32_t*)setup->sin_s)[0]); source_ptr += strlen(source_buf + source_ptr);
    for (i = 1; i < num_proj; i++) {
	sprintf(source_buf + source_ptr, ",0x%08x", ((uint32_t*)setup->sin_s)[i]); source_ptr += 11;
    }
    sprintf(source_buf + source_ptr, "};\n"); source_ptr += strlen(source_buf + source_ptr);

    sprintf(source_buf + source_ptr, "__constant unsigned axis_itable[] = {0x%08x", ((uint32_t*)setup->axis_position_corr_s)[0]); source_ptr += strlen(source_buf + source_ptr);
    for (i = 1; i < num_proj; i++) {
	sprintf(source_buf + source_ptr, ",0x%08x", ((uint32_t*)setup->axis_position_corr_s)[i]); source_ptr += 11;
    }
    sprintf(source_buf + source_ptr, "};\n"); source_ptr += strlen(source_buf + source_ptr);
    memcpy(source_buf + source_ptr, source, source_len + 1);
    //source_buf[source_ptr] = 0;    
    //puts(source_buf);

    ctx->program = clCreateProgramWithSource(ctx->clctx, 1, (const char**)&source_buf, NULL, &err);
    
    free(source_buf);
#else /* HST_OPENCL_CONST_HACK */
    const char *source_end = strstr(source, "//HST_OPENCL_CONST_HACK");
    if (source_end) {
	size_t source_len = source_end - source;
        char *source_buf = malloc(source_len + 1);
	check_alloc(source_buf, "source_buf");
	memcpy(source_buf, source, source_len);
	source_buf[source_len] = 0;
        
	ctx->program = clCreateProgramWithSource(ctx->clctx, 1, (const char**)&source_buf, NULL, &err);
	free(source_buf);
    } else {
	ctx->program = clCreateProgramWithSource(ctx->clctx, 1, &source, NULL, &err);
    }
# endif /* HST_OPENCL_CONST_HACK */
    CL_CHECK_ERROR(err);
#endif /* HST_OPENCL_LOAD_PTX */    

    const char *build_flags = HST_OPENCL_BUILD_FLAGS;

    err = clBuildProgram(ctx->program, 0, NULL, build_flags, NULL, NULL);
    if (err != CL_SUCCESS) {
        const int LOG_SIZE = 4096;
        char* log = (char*) malloc(LOG_SIZE * sizeof(char));
        CL_CHECK_ERROR_EX(clGetProgramBuildInfo(ctx->program, ctx->device_id, CL_PROGRAM_BUILD_LOG, LOG_SIZE, (void*) log, NULL));
        pyhst_error("OpenCL Build Log\n---------------\n%s", log);
        free(log);
        return 255;
    }

#ifdef HST_OPENCL_SAVE_PTX    
    void *binary;
    size_t binary_size;
    size_t real_size;

    err = clGetProgramInfo(ctx->program, CL_PROGRAM_BINARY_SIZES, sizeof(size_t), &binary_size, &real_size);
    if (err == CL_SUCCESS) binary = malloc(binary_size);
    if ((binary)&&(err == CL_SUCCESS)) {
	FILE *f;
	err = clGetProgramInfo(ctx->program, CL_PROGRAM_BINARIES, sizeof(void*), &binary, &real_size);
	if (err == CL_SUCCESS) f = fopen(HST_OPENCL_SAVE_PTX, "wb");
	else f = NULL;
	
	if (f) {
	    (void)fwrite(binary, 1, binary_size, f);
	    fclose(f);
	}
	free(binary);
    }
#endif /* HST_OPENCL_SAVE_PTX */

    for (kernel_id = 0; kernel_id < HST_KERNEL_NUMBER; kernel_id++) {
	const char *kernel_name = hst_opencl_resolve_kernel(ctx, kernel_id, variants?variants[kernel_id]:0);
	printf("%s: loading the kernel %s\n", ctx->device_name, kernel_name);
	ctx->kernel[kernel_id] = clCreateKernel(ctx->program, kernel_name, &err);
        CL_CHECK_ERROR_EX(err);
        if (err != CL_SUCCESS) {
            pyhst_error("Undefined kernel name %s", kernel_name);
	    continue;
	}
	
	const HSTKernelDescription *d = hst_opencl_resolve_kernel_description(ctx, kernel_name);
	if ((d)&&(d->shared_memory)) {
	    cl_uint args;
	    err = clGetKernelInfo(ctx->kernel[kernel_id], CL_KERNEL_NUM_ARGS, sizeof(cl_uint), &args, NULL);
	    if (err) {
	        pyhst_error("Unable to determine number of arguments for kernel %s", kernel_name);
		continue;
	    }
	    
	    err = clSetKernelArg(ctx->kernel[kernel_id], args - 1, d->shared_memory, NULL);
	    if (err) {
	        pyhst_error("Unable to set shared memory argument for kernel %s", kernel_name);
		continue;
	    }
/*	    
	    if (!strcmp(kernel_name, "hst_backproject_direct_oversample4")) {
	        err = clSetKernelArg(ctx->kernel[kernel_id], args - 2, d->shared_memory, NULL);
	    }

	    if (err) {
	        pyhst_error("Unable to set shared memory argument for kernel %s", kernel_name);
		continue;
	    }
*/
	}
    }
    
    return err;
}
