/*
 * The PyHST program is Copyright (C) 2002-2011 of the
 * European Synchrotron Radiation Facility (ESRF) and
 * Karlsruhe Institute of Technology (KIT).
 *
 * PyHST is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * hst is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
struct {
    cl_platform_id platform_id;
    cl_uint num_devices;
    cl_device_id device_ids[HST_MAX_OPENCL_DEVICES]; 
} g_hst_platform;

//static HSTConstString hst_opencl_timers[] = { "complete", "data transfer", "filtering", "backprojection", NULL }; 
static HSTConstString hst_opencl_timers[] = { 
    "complete reconstruction",
    "transfer to device", "transfer from device", "texture mapping", "projection filtering", "backprojection", "*initialization and cleanup", 
    "OpenCL complete", "OpenCL to device", "OpenCL from device",  "OpenCL texture", "OpenCL filtering", "OpenCL backprojection", "*OpenCL init",
    NULL
};



#ifdef PYHST_MEASURE_TIMINGS
inline void hst_measure_time(cl_ulong *timer, cl_command_queue queue, cl_event event)
{
    cl_ulong start, end;
    if (queue) clFinish(queue);
    CL_CHECK_ERROR(clGetEventProfilingInfo(event, CL_PROFILING_COMMAND_END,
            sizeof(cl_ulong), &end, NULL));
    CL_CHECK_ERROR(clGetEventProfilingInfo(event, CL_PROFILING_COMMAND_START,
            sizeof(cl_ulong), &start, NULL));
    *timer += (end - start);
}
#else
#define hst_measure_time(timer, queue, event)
#endif

static HSTConstString *hst_opencl_get_timers(HSTReconstructorConstContextPtr rctx, double *timers)
{
#ifdef PYHST_MEASURE_TIMINGS
    int i;
    int report = 0;
    assert(rctx);

    if (timers) {
        OpenCLContext *ctx = (OpenCLContext *) rctx;
        size_t timer_resolution;
        CL_CHECK_ERROR(clGetDeviceInfo(ctx->device_id,
                    CL_DEVICE_PROFILING_TIMER_RESOLUTION,       //in nanoseconds 
                    sizeof(size_t), &timer_resolution, NULL));

        timers[1] = g_timer_elapsed(ctx->togpu_timer, NULL);
        timers[2] = g_timer_elapsed(ctx->fromgpu_timer, NULL);
        timers[3] = g_timer_elapsed(ctx->texture_timer, NULL);
        timers[4] = g_timer_elapsed(ctx->pre_timer, NULL);
        timers[5] = g_timer_elapsed(ctx->main_timer, NULL);
        timers[6] = g_timer_elapsed(ctx->init_timer, NULL);
        timers[0] = timers[1] + timers[2] + timers[3] + timers[4] + timers[5];

        timers[8] = ctx->timer_togpu * 1.0e-9;
        timers[9] = ctx->timer_fromgpu * 1.0e-9;
        timers[10] = ctx->timer_texture * 1.0e-9;
        timers[11] = ctx->timer_filter * 1.0e-9;
        timers[12] = ctx->timer_backprojection * 1.0e-9;
        timers[13] = ctx->timer_init * 1.0e-9;
        
        for (i = 8; i < 14; i++) {
	    if (timers[i] > 10000) {
	        report = 1;
	        timers[i] = 0;	// Some bug in Apple FFT library
	    }
	}
	
	if (report) {
	    pyhst_warning("Problem with OpenCL timers...");
        }
        
        timers[7] = timers[8] + timers[9] + timers[10] + timers[11] + timers[12];
    }
#endif
    
    return hst_opencl_timers;
}
